package com.example.micrometer;

import io.micrometer.observation.ObservationRegistry;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.RestController;
import reactor.core.publisher.Mono;
import reactor.core.scheduler.Scheduler;

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutorService;

import static java.util.concurrent.CompletableFuture.supplyAsync;
import static reactor.core.observability.micrometer.Micrometer.observation;
import static reactor.core.publisher.Mono.deferContextual;
import static reactor.core.publisher.Mono.fromCompletionStage;

@RestController
public class ExampleController {
    private final ExecutorService ioExecutorService;
    private final ObservationRegistry registry;
    private final Scheduler customScheduler;


    public ExampleController(ObservationRegistry registry, ExecutorService ioExecutorService, Scheduler customScheduler) {
        this.registry = registry;
        this.ioExecutorService = ioExecutorService;
        this.customScheduler = customScheduler;

    }

    private Mono<String> getCache() {
        return observe("debugme", delayedStringMono("a")
                .flatMap(a -> observe("inner", delayedStringMono(a + "b")))
        );
    }

    public <T> Mono<T> observe(String name, Mono<T> mono) {
        return deferContextual(view -> mono.publishOn(customScheduler).contextWrite(view).name(name)
                .tap(observation(registry)).subscribeOn(customScheduler));
    }

    @GetMapping("/example")
    public Mono<ResponseEntity<String>> example() {
        CompletableFuture<String> completionStage = getCache().toFuture();
        return fromCompletionStage(completionStage).map(content -> ResponseEntity.ok().body(content));
    }

    private Mono<String> delayedStringMono(String start) {
        return fromCompletionStage(() -> supplyAsync(() -> delayedString(start), ioExecutorService));
    }

    private static String delayedString(String string) {
        try {
            Thread.sleep(1000);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
        return string;

    }
}
